/*
 * Author: John Huddleston
 * Purpose: Create a gridded shapefile
 */
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <getopt.h>
#include <signal.h>
#include <proj_api.h>
double atof();
int pnt2grid(char *,char *,float,int,int,float **,float **,double **,double **);
 
const char *grid2shp_id(void)
{
return("@(#) $Id: grid2shp.c 348 2011-04-22 07:27:45Z jhudd $");
}

void grid2shp_version(void)
{
  printf("The version is 20110418 1.0\n");
}

void grid2shp_usage(void)
{
  printf("Usage: -V          ==> version, date information\n");
  printf("Usage: -u          ==> usage, this information\n");
  printf("Usage: -h          ==> help, needs more help at this point\n");
  printf("Usage: -1 XXX      ==> set lower left X coord, default is -2736000\n");
  printf("Usage: -2 YYY      ==> set lower left Y coord, default is -2088000\n");
  printf("Usage: -x XXX      ==> set the number of X grids, default is 148\n");
  printf("Usage: -y YYY      ==> set the number of Y grids, default is 112\n");
  printf("Usage: -s sz       ==> set the grid size value, default is 36000\n");
  printf("Usage: -p          ==> position to lower left grid corner, default is no\n");
  printf("Usage: -o file     ==> set the output polygon shapefile\n");
}

void grid2shp_help(void)
{
  printf("The program will process point data into a gridded dataset\n");
  printf("If the User specifies the -p option, the lower left point is printed.\n");
}
int grid2shp(char *shpName,float sz,float lx,float ly,float rx,float uy, int pos)
{
	int ch, i,j,r;
	int xpos,ypos;
	int loopcnt,cnt;
	float v,ew,ns, xdist, ydist;
	char  chr, *p, *q, inbuf[256];
	char  hdrName[128];
	float **P1, **P2;
	double **Lon, **Lat;
	int pret,szx,szy;
	projPJ pj_latlong, pj_lcc;
	szx = (int)((rx-lx)/sz); /* number of columns */
	szy = (int)((uy-ly)/sz); /* number of rows */
	if((szx % 2) == 1) szx++;
	if((szy % 2) == 1) szy++;
	/* row major order ==> allocate number of rows 
	 * then allocate the floats in the columns
	 */
	P1 = malloc(szx*sizeof(float *)); /* X position */
	P2 = malloc(szx*sizeof(float *)); /* Y position */
	Lon = malloc(szx*sizeof(double *)); /* Lon matrix values */
	Lat = malloc(szx*sizeof(double *)); /* Lat matrix output grid */
	for(i=0; i<szx; i++) /* e.g. szx=70 */
	{
		P1[i] = malloc(szy*sizeof(float));
		P2[i] = malloc(szy*sizeof(float));
		Lon[i] = malloc(szy*sizeof(double));
		Lat[i] = malloc(szy*sizeof(double));
	}
	if(pos != 0)
	{
		for(i=0; i<szx; i++)
		{
			for(j=0; j<szy; j++)
			{
				P1[i][j] = (float)((lx+i*sz+(sz/2.0)));
				Lon[i][j]=P1[i][j];
				P2[i][j] = (float)((ly+j*sz+(sz/2.0)));
				Lat[i][j]=P2[i][j];
			}
		}
	}
	else
	{
		for(i=0; i<szx; i++)
		{
			for(j=0; j<szy; j++)
			{
				P1[i][j] = (float)((lx+i*sz));
				Lon[i][j]=P1[i][j];
				P2[i][j] = (float)((ly+j*sz));
				Lat[i][j]=P2[i][j];
			}
		}
	}
	pj_latlong = pj_init_plus("+proj=latlong +datum=WGS84");
	if (!(pj_latlong)) {
		fprintf(stderr,"error creating NAD83 datum");
		return(20);
	}
	pj_lcc = pj_init_plus("+proj=lcc +lat_1=45 +lat_2=33 +lat_0=40 +lon_0=-97 +x_0=0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
	if (!(pj_lcc) ) {
		pj_free(pj_latlong);
		fprintf(stderr,"error creating nad83 lcc");
		return(30);
	}
	for(i=0; i<szx; i++)
	{
		for(j=0; j<szy; j++)
		{
			pret = pj_transform(pj_lcc, pj_latlong, 1, 1, &Lon[i][j], &Lat[i][j], NULL );
			Lon[i][j] /= DEG_TO_RAD; Lat[i][j] /= DEG_TO_RAD;
		}
	}
	pj_free(pj_latlong);
	pj_free(pj_lcc);
	sprintf(hdrName,"X float(24,11)\tY float(24,11)\tLongitude double(24,11)\tLatitude double(24,11)");
	pnt2grid(shpName,hdrName,sz,szx,szy,P1,P2,Lon,Lat);

	/* free up resources */
	for(i=0; i<szx; i++)
	{
		free(Lon[i]);
		free(Lat[i]);
		free(P1[i]);
		free(P2[i]);
	}
	free(Lon);
	free(Lat);
	free(P1);
	free(P2);
	return(0);
}

int main(argc,argv)
int argc;
char *argv[];
{
  int i;
  float lx = -2736000.0; /* left X position, use -1 option to change it */
  float ly = -2088000.0;   /* lower Y position, use -2 option to change it */
  float rx = 2592000.0; /* right X position, use -x option to change it */
  float uy = 1944000.0;   /* upper Y position, use -y option to change it */
  float sz = 36000.;   /* default size of grid */
  int   xGrids=148;
  int   yGrids=112;
  int position = 1; /* center of each grid */
  int isatty();
  int iret, atoi();
  char shpName[64];
  sprintf(shpName,"%s","Grid2SHP");

  if (argc == 1)
  {
	iret = grid2shp(shpName,sz,lx,ly,rx,uy,position);
    return(iret);
  }
  else
  {
    while ((i = getopt(argc, argv, "Vuhpo:s:1:2:x:y:")) != -1) {
      switch (i) {
      case 'V' :
        grid2shp_version();
        return(0);
        break;
      case 'u' :
        grid2shp_usage();
        return(0);
        break;
      case 'h' :
        grid2shp_help();
        return(0);
        break;
      case 'o' :
		sprintf(shpName,"%s",optarg);
        break;
      case 's' :
        sz = (float)atof(optarg);
        break;
      case '1' :
        lx = (float)atof(optarg);
        break;
      case '2' :
        ly = (float)atof(optarg);
        break;
      case 'x' :
        xGrids = (float)atof(optarg);
        break;
      case 'y' :
        yGrids = (float)atof(optarg);
        break;
      case 'p' :
        position = 0; /* the lower left grid */
        break;
      }
    }
  }
  rx=lx+(sz*xGrids);
  uy=ly+(sz*yGrids);
  iret = grid2shp(shpName,sz,lx,ly,rx,uy,position);
  return(iret);
}
