#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shapefil.h"

#define STR_BUFFER_SIZE		128
double sqrt(), cos(), sin();
int atoi();
#define MY_PI                3.1415926535897932384
int point2square(double n, double r, float lat, float lon, double *v, double *w)
{ /* e.g. 4, radius, Y, X, &xPoly[0], &yPoly[0] */
    int i, s;
    double d, x, y;
    s = (int)((n * 2) - 1); /* starting number, e.g. s = 7 if n = 4 */
    for (i = s; i > 0; i-=2)
    {
        d = i * MY_PI / n;
        x = lon + r * cos(d); *v++ = x;
        y = lat + r * sin(d); *w++ = y;
    }
    d = s * MY_PI / n;
    x = lon + r * cos(d); *v = x;
    y = lat + r * sin(d); *w = y;
    return (0);
}
static void WritePolygon(SHPHandle hSHP, int rec, int coords, double * x, double * y, int nparts, int * partstarts)
{
	SHPObject   *psShape;
	psShape = SHPCreateObject( SHPT_POLYGON, rec, nparts, partstarts, NULL, coords, x, y, NULL, NULL );
	SHPWriteObject( hSHP, -1, psShape );
	SHPDestroyObject( psShape );
}

static int MakeAttributes (char *linebuf, DBFHandle hDBF)
{
	int chr;
	char		*p, *q, *r, *s;
	char		fieldname[STR_BUFFER_SIZE];
/* example
X double(24,11) Y double(24,11) Value double(24,11) Estimate double(24,11)
*/
	if(linebuf == NULL) return(-1);
	s = &linebuf[0];
	while(*s != 0)
	{
		p = s;
		while(*p == ' ') p++;
		q = p;
		while(*p != ' ' && *p != '\t' && *p != ',' && *p != 0) p++; 
		if(*p != 0) *p++=0;
		sprintf(fieldname, "%s", q);
		while(*p == ' ' || *p == '\t' || *p == ',') p++;
		chr = *p; /* char(30) */
		while(*p != '(' && *p != 0) p++; if(*p != 0) *p++=0;

		q = p; while(*q != ')' && *q != 0 && *q != ',') q++;
		if(*q == ')')
		{
			*q++ = 0; /* e.g. int(11) */
			s = q;
		}
		else if(*q == ',')
		{
			*q++ = 0; /* e.g. double (11,2) */
			r = q; while(*r != ')' && *r != 0) r++;
			if(*r != 0) *r++ = 0; s = r;
		}
		else
			s = q;
		while(*s == ' ' || *s == '\t' || *s == ',') s++;
		switch(chr)
		{
		/* e.g. char(11) */
		case 'c':
		case 'C':
		case 's': /* string */
		case 'S':
			if( DBFAddField( hDBF, fieldname, FTString, atoi(p), 0 ) == -1 )
			{
				fprintf(stderr,"DBFAddField(%s,FTString,%d,0) failed.\n", fieldname, atoi(p) );
				return( -2 );
			}
			break;
		/* e.g. int(11) or integer(11) */
		case 'i':
		case 'I': /* if loading into ArcSDE use the smallint */
			if( DBFAddField( hDBF, fieldname, FTInteger, atoi(p), 0 ) == -1 )
			{
				fprintf(stderr,"DBFAddField(%s,FTInteger,%d,0) failed.\n", fieldname, atoi(p) );
				return( -3 );
			}
			break;
		/* e.g. float(11,2) or double(11,2) */
		case 'd':
		case 'f':
		case 'D':
		case 'F':
			if( DBFAddField( hDBF, fieldname, FTDouble, atoi(p), atoi(q) ) == -1 )
			{
				fprintf(stderr,"DBFAddField(%s,FTDouble,%d,%d) failed.\n", fieldname, atoi(p), atoi(q) );
				return( -4 );
			}
			break;
		}
	}
	return(0);
}
int pnt2grid(char *shpName,char *header,float sz,int sizeX,int sizeY,float **X,float **Y,double **P,double **Z)
{
	DBFHandle hDBF;		/* handle for dBase file */
	SHPHandle hSHP;		/* handle for shape files .shx and .shp */
	FILE *fp;
	int iret;
	char		dbffname[STR_BUFFER_SIZE];
	char		shpfname[STR_BUFFER_SIZE];
	char		prjfname[STR_BUFFER_SIZE];
	int r,x,y;
	int rec = 0;
	int coord, nparts;
	int partstarts[1];
	double  radius,xcoord[5],ycoord[5];
	radius = sqrt(sz*sz/4.0)*sqrt(2.0);

	if(shpName == NULL) return(-1);
	sprintf(dbffname, "%s.dbf", shpName);

	hDBF = DBFCreate( dbffname );
	if( hDBF == NULL ) {
		fprintf(stderr, "DBFCreate(%s) failed.\n", dbffname );
		return(-2);
	}
	iret = MakeAttributes(header, hDBF);
	if(iret < 0)
	{
		fprintf(stderr, "MakeAttributes(%s) failed.\n", shpName );
		DBFClose( hDBF );
		return(-3);
	}
	DBFClose( hDBF );
	hDBF = DBFOpen( dbffname, "r+b" );
	if( hDBF == NULL ) {
		fprintf(stderr, "DBFOpen(%s,\"r+b\") failed.\n", dbffname );
		return(-4);
	}

	sprintf(shpfname, "%s.shp", shpName);
	hSHP = SHPCreate( shpfname, SHPT_POLYGON );
	if( hSHP == NULL ) {
		fprintf(stderr, "SHPOpen(%s, SHPT_POLYGON) failed.\n", shpfname );
		DBFClose( hDBF );
		return(-5);
	}
	coord = 5;
	nparts = 0;
	partstarts[0] = 0;
	for(x = 0; x < sizeX; x++)
	{
		for(y = 0; y < sizeY; y++)
		{
			DBFWriteDoubleAttribute(hDBF, rec, 0, (double)X[x][y]);
			DBFWriteDoubleAttribute(hDBF, rec, 1, (double)Y[x][y]);
			DBFWriteDoubleAttribute(hDBF, rec, 2, (double)P[x][y]);
			DBFWriteDoubleAttribute(hDBF, rec, 3, (double)Z[x][y]);
			r = point2square(4,radius,Y[x][y],X[x][y],&xcoord[0],&ycoord[0]);
			WritePolygon(hSHP,rec,coord,&xcoord[0],&ycoord[0],nparts,&partstarts[0]);
			rec++;
		}
	}

	/* Finish output files */
	DBFClose( hDBF );
	SHPClose( hSHP );

	sprintf(prjfname, "%s.prj", shpName);
	fp = fopen(prjfname,"w+");
	if(fp != NULL) {
    fprintf(fp,"PROJCS[\"North_America_Lambert_Conformal_Conic\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137,298.257222101]],PRIMEM[\"Greenwich\",0],UNIT[\"Degree\",0.0174532925199432955]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0],PARAMETER[\"False_Northing\",0],PARAMETER[\"Central_Meridian\",-96],PARAMETER[\"Standard_Parallel_1\",20],PARAMETER[\"Standard_Parallel_2\",60],PARAMETER[\"Latitude_Of_Origin\",40],UNIT[\"Meter\",1]]");
	  fclose(fp);
	}
	/* success */
	return(0);
}
